#pragma once 
#include "hmbdc/Compile.hpp"

#include <type_traits>
#include <tuple>

namespace hmbdc {
template <typename T>
struct function_traits: function_traits<decltype(&T::operator())>
{};

template <typename ClassType, typename ReturnType, typename ...Args>
struct function_traits<ReturnType(ClassType::*)(Args...) const> {
    enum { arity = sizeof...(Args) };
    typedef ReturnType result_type;

    template <size_t i>
    struct arg {
        using type = typename std::tuple_element<i, std::tuple<Args...>>::type;
    };
};

template <typename ClassType, typename ReturnType, typename ...Args>
struct function_traits<ReturnType(ClassType::*)(Args...)> {
    enum { arity = sizeof...(Args) };

    typedef ReturnType result_type;

    template <size_t i>
    struct arg {
        using type = typename std::tuple_element<i, std::tuple<Args...>>::type;
    };
};

template <typename Base, typename Derived>
struct is_first_base_of{
    enum {
        value = std::is_base_of<Base, Derived>::value 
               //&& (Base const*)((Derived const*)0xffff) == ((Base const*)0xffff) - not supported on some compilers
    };
};

template <typename T, typename Tuple>
struct index_in_tuple;

template <typename T, typename ...Types>
struct index_in_tuple<T, std::tuple<T, Types...>> {
    static const std::size_t value = 0;
};

template <typename T>
struct index_in_tuple<T, std::tuple<>> {
    static const std::size_t value = 0;
};

template <typename T, typename U, typename ...Types>
struct index_in_tuple<T, std::tuple<U, Types...>> {
    static const std::size_t value = 1 + index_in_tuple<T, std::tuple<Types...>>::value;
};


template <typename T, typename Tuple>
struct add_if_not_in_tuple {
    using type = std::tuple<T>;
};

template <typename T, typename ...Types>
struct add_if_not_in_tuple<T, std::tuple<Types ...>> {
private:
    using Tuple = std::tuple<Types ...>;
public:    
    using type = typename std::conditional<
        index_in_tuple<T, Tuple>::value == sizeof...(Types)
        , std::tuple<Types ..., T>
        , Tuple
        >::type;
};

template <typename Tuple1, typename Tuple2>
struct merge_tuple_unique;

template <typename Tuple1>
struct merge_tuple_unique<Tuple1, std::tuple<>> {
    using type = Tuple1;
};

template <typename Tuple1, typename T, typename ...Types>
struct merge_tuple_unique<Tuple1, std::tuple<T, Types...>> {
private:
    using step1 = typename add_if_not_in_tuple<T, Tuple1>::type;
public:
    using type = typename merge_tuple_unique<step1, std::tuple<Types...>>::type;
};

template <typename Tuple>
struct max_size_in_tuple {
    static const std::size_t value = 0;
};

template <typename T, typename ...Ts>
struct max_size_in_tuple<std::tuple<T, Ts...>> {
private:
    static const std::size_t step1 = max_size_in_tuple<std::tuple<Ts...>>::value;
public:
    static const std::size_t value = step1 > sizeof(T) ? step1 : sizeof(T);
};
template <typename Tuple0, typename Tuple1>
struct concat_tuple;

template <typename ...T, typename ...U>
struct concat_tuple<std::tuple<T...>, std::tuple<U...>> {
    using type = std::tuple<T..., U...>;
};

template <template <class, class> class pred, typename T, typename ...Ts>
struct insert_in_ordered_tuple {
    using type = std::tuple<T>;
};
template <template <class, class> class pred, typename T
    , typename T0, typename ...Ts>
struct insert_in_ordered_tuple<pred, T, std::tuple<T0, Ts...>> {
    using type = typename std::conditional<
        pred<T, T0>::value
        , std::tuple<T, T0, Ts...>
        , typename concat_tuple<
            std::tuple<T0>
            , typename insert_in_ordered_tuple<
                pred, T, std::tuple<Ts...>
            >::type
        >::type
    >::type;
};

template <template <class, class> class pred, typename Tuple>
struct sort_tuple;

template <template <class, class> class pred>
struct sort_tuple<pred, std::tuple<>> {
    using type = std::tuple<>;
};

template <template <class, class> class pred, typename T, typename ...Ts>
struct sort_tuple<pred, std::tuple<T, Ts...>> {
    using type = typename insert_in_ordered_tuple<
        pred
        , T
        , typename sort_tuple<pred, std::tuple<Ts...>>::type
    >::type;
};

#if __GNUC_PREREQ(9,0)
template <typename Tuple, size_t from = 0, size_t to = std::tuple_size<Tuple>::value>
struct bsearch_tuple;

template <typename ...Ts, size_t from, size_t to>
struct bsearch_tuple<std::tuple<Ts...>, from, to> {
    template <typename comp>
    bool operator()(comp&& c) {
        if constexpr (from >= to) {
            return false;
        } else {
            constexpr auto at = (from + to) / 2;
            using atType = typename std::tuple_element<at, std::tuple<Ts...>>::type;
            int comp_res = 
                c((atType*)nullptr);
            if (comp_res == 0) 
                return true;
            else if (comp_res < 0) 
                return bsearch_tuple<std::tuple<Ts...>, from, at>()(
                    std::forward<comp>(c));
            else
                return bsearch_tuple<std::tuple<Ts...>, at + 1, to>()(
                    std::forward<comp>(c));
        }
    }
};

template <template <class> class target_template, typename M>
constexpr bool is_template = false;

template <template <class> class target_template, typename N>
constexpr bool is_template<target_template, target_template<N>> = true;

#endif

template <template <class> class target_template, typename Tuple>
struct templatized_aggregator {
    using type = std::tuple<>;
};

template <template <class> class target_template, typename T, typename ...Ts>
struct templatized_aggregator<target_template, std::tuple<T, Ts...>> {
    using type = 
        typename templatized_aggregator<target_template, std::tuple<Ts...>>::type;
};

template <template <class> class target_template, typename T, typename ...Ts>
struct templatized_aggregator<target_template, std::tuple<target_template<T>, Ts...>> {
    using type = typename merge_tuple_unique<std::tuple<target_template<T>>
        , typename templatized_aggregator<target_template, std::tuple<Ts...>>::type
    >::type;
};

template <template <class> class target_template, typename Tuple>
struct templatized_subtractor {
    using type = std::tuple<>;
};

template <template <class> class target_template, typename T, typename ...Ts>
struct templatized_subtractor<target_template, std::tuple<T, Ts...>> {
    using type = typename merge_tuple_unique<std::tuple<T>
        , typename templatized_subtractor<target_template, std::tuple<Ts...>>::type
    >::type;
};

template <template <class> class target_template, typename T, typename ...Ts>
struct templatized_subtractor<target_template, std::tuple<target_template<T>, Ts...>> {
    using type = typename templatized_subtractor<target_template, std::tuple<Ts...>>::type;
};

template<typename T, size_t sz>
size_t length_of(T (&)[sz]) { return sz; }
}
