#include "hmbdc/Copyright.hpp"
#pragma once

#include "hmbdc/app/Message.hpp"
#include "hmbdc/app/Logger.hpp"
#include "hmbdc/MetaUtils.hpp"

namespace hmbdc { namespace app {

namespace messagehandler_detail {
template <typename CcClient, typename Message>
struct dispatcher {
    bool operator()(CcClient& c, MessageHead& m) {
        if (m.typeTag == Message::typeTag) {
            HMBDC_LOG_r(c.hmbdcName(), this, " received ", m.get<Message>());
            c.handleMessageCb(m.get<Message>());
            using Wrap = MessageWrap<Message>;
            auto& wrap = static_cast<Wrap&>(m);
            wrap.~Wrap();
            return true;
        }
        return false;
    }
};

template <typename CcClient>
struct dispatcher<CcClient, JustBytes> {
    bool operator()(CcClient& c, MessageHead& m) {
        if (m.typeTag > 1000) { //only deliver user messages
            auto j = &m.get<uint8_t>();
            HMBDC_LOG_r(c, this, " received bytes typeTag=", m.typeTag);
            c.handleJustBytesCb(m.typeTag, j);
        }
        return true;
    }
};

template<class T, class U>
struct ascending_by_tag 
: std::integral_constant<bool, T::typeTag < U::typeTag>
{};


template <typename CcClient, typename ... Messages>
struct SeqMessageDispacher {
    bool operator()(CcClient&, MessageHead&) const{return false;}
};

template <typename CcClient, typename M, typename ... Messages>
struct SeqMessageDispacher<CcClient, std::tuple<M, Messages...>> {
    bool operator()(CcClient& c, MessageHead& e) const {
        if (!dispatcher<CcClient, M>()(c, e))
            return SeqMessageDispacher<CcClient, std::tuple<Messages ...>>()(c, e);
        else
            return true;
    }
};

}

#if __GNUC_PREREQ(9,0)
template <typename CcClient, typename ... Messages>
struct MessageDispacher {
    bool operator()(CcClient&, MessageHead&) const;
};

template <typename CcClient, typename ... Messages>
struct MessageDispacher<CcClient, std::tuple<Messages...>> {
    static_assert(
        index_in_tuple<JustBytes, std::tuple<Messages...>>::value == sizeof...(Messages)
        || sizeof...(Messages) <= 16
        , "JustBytes cannot be combined with too many other types");

    bool operator()(CcClient& c, MessageHead& e) const {
        if constexpr (sizeof...(Messages) > 16) {
            using SortedTupleMessages
                = typename sort_tuple<messagehandler_detail::ascending_by_tag
                    , std::tuple<Messages...>>::type;

            return bsearch_tuple<SortedTupleMessages>()(
                [&c, &e](auto* t) {
                    using M = typename std::remove_reference<decltype(*t)>::type;
                    if (messagehandler_detail::dispatcher<CcClient, M>()(c, e)) {
                        return 0;
                    }
                    if (e.typeTag < M::typeTag) return -1;
                    return 1;
                }
            );
        } else {
            return messagehandler_detail::SeqMessageDispacher<
                CcClient, std::tuple<Messages ...>>()(c, e);
        }
    }
};

#else

template <typename CcClient, typename ... Messages>
using MessageDispacher = messagehandler_detail::SeqMessageDispacher<
    CcClient, Messages ...>;
#endif

}}
